// (C) Copyright 2015 Martin Dougiamas
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

import { Component, Input, OnInit, OnChanges, SimpleChange } from '@angular/core';
import { NavController } from 'ionic-angular';
import { CoreSitesProvider } from '@providers/sites';
import { CoreUtilsProvider } from '@providers/utils/utils';

/**
 * Component to display a "user avatar".
 *
 * Example: <ion-avatar core-user-avatar [user]="participant"></ion-avatar>
 */
@Component({
    selector: 'ion-avatar[core-user-avatar]',
    templateUrl: 'core-user-avatar.html'
})
export class CoreUserAvatarComponent implements OnInit, OnChanges {
    @Input() user: any;
    // The following params will override the ones in user object.
    @Input() profileUrl?: string;
    @Input() protected linkProfile = true; // Avoid linking to the profile if wanted.
    @Input() fullname?: string;
    @Input() protected userId?: number; // If provided or found it will be used to link the image to the profile.
    @Input() protected courseId?: number;
    @Input() checkOnline = false; // If want to check and show online status.

    // Variable to check if we consider this user online or not.
    // @TODO: Use setting when available (see MDL-63972) so we can use site setting.
    protected timetoshowusers = 300000; // Miliseconds default.
    protected myUser = false;
    protected currentUserId: number;

    constructor(private navCtrl: NavController, private sitesProvider: CoreSitesProvider, private utils: CoreUtilsProvider) {
        this.currentUserId = this.sitesProvider.getCurrentSiteUserId();
    }

    /**
     * Component being initialized.
     */
    ngOnInit(): void {
        this.setFields();
    }

    /**
     * Listen to changes.
     */
    ngOnChanges(changes: { [name: string]: SimpleChange }): void {
        // If something change, update the fields.
        if (changes) {
            this.setFields();
        }
    }

    /**
     * Set fields from user.
     */
    protected setFields(): void {
        this.profileUrl = this.profileUrl || (this.user && (this.user.profileimageurl || this.user.userprofileimageurl ||
            this.user.userpictureurl || this.user.profileimageurlsmall));

        if (typeof this.profileUrl != 'string') {
            this.profileUrl = '';
        }

        this.fullname = this.fullname || (this.user && (this.user.fullname || this.user.userfullname));

        this.userId = this.userId || (this.user && (this.user.userid || this.user.id));
        this.courseId = this.courseId || (this.user && this.user.courseid);

        // If not available we cannot ensure the avatar is from the current user.
        this.myUser = this.userId && this.userId == this.currentUserId;

    }

    /**
     * Helper function for checking the time meets the 'online' condition.
     *
     * @return boolean
     */
    isOnline(): boolean {
        if (this.myUser || this.utils.isFalseOrZero(this.user.isonline)) {
            return false;
        }

        if (this.user.lastaccess) {
            // If the time has passed, don't show the online status.
            const time = new Date().getTime() - this.timetoshowusers;

            return this.user.lastaccess * 1000 >= time;
        } else {
            return this.user.isonline;
        }
    }

    /**
     * Function executed image clicked.
     */
    gotoProfile(event: any): void {
        if (this.linkProfile && this.userId) {
            event.preventDefault();
            event.stopPropagation();
            this.navCtrl.push('CoreUserProfilePage', { userId: this.userId, courseId: this.courseId });
        }
    }
}
